# coding=utf-8

"""Utilities for Code_Aster hooks and extension."""

import os
import re
from subprocess import Popen, PIPE
import getpass
import smtplib
import time
from collections import OrderedDict

from mercurial import hg, commands, ui as UI
from aslint.i18n import _
from aslint.config import ASCFG
from aslint.logger import logger, magenta, blue, cyan, green, red, grey
from aslint.string_utils import to_unicode, convert

RE_ISSUE = r'(?P<issue>(?P<type>\#|issue|Issue|bb|BB)(?: ?)(?P<number>[0-9]{4,6}))'
BUGTRACKERS = (None, 'rex', 'bitbucket')
# closed issues must be enclosed by '[]': [#1234]

# identifier of the repositories (hash of revision 0 or of the revision that
# opens the branch for derivated repository)
REPOID = OrderedDict([
    ('c6b122a7a53ea4f82f2e4b198e1357f426d66f71', 'data-xx'),
    ('08ee4973b18cc37bc89a3ed356b91397a352b9c1', 'data'),
    ('852b9b2bfbed59991756733f65ca422a08df5383', 'src-xx'),
    ('becdb3e46d91de04ce69a240b714ec9df3a50826', 'src'),
    ('54d590227bd7012028615e8be14866b8e22174e3', 'validation-xx'),
    ('8918a03f384fdd666260a381d5dad564ec0a177d', 'validation'),
    ('d9ffe8b5560b5799cf9821b9c08d0f6f41497dee', 'devtools'),
    ('ee137aec1fe656351ca17b42751f2a0de5f157f0', 'salome-codeaster-study'),
    ('550c13b26fd780e034e54d79200f5ec7511456a4', 'salome-meca-rtool'),
])


def is_admin(user):
    """Tell if 'user' is admin."""
    return re.search(r'^["\']?(code_aster|mergerobot)', user, re.I) is not None


def shortrev_s(rev):
    """Return a short representation of a revision passed as string"""
    return rev[:16]


def shortrev(rev):
    """Return a short representation of a revision"""
    return shortrev_s(rev.hex())


def hgcmd(ui, cmd, repo, *args, **kwargs):
    """Return the output of `cmd`."""
    if 'message' in kwargs:
        kwargs['message'] = convert(kwargs['message'])
    ui.pushbuffer()
    iret = cmd(ui, repo, *args, **kwargs)
    out = ui.popbuffer()
    return iret, out.strip()


def newui():
    """Return a ui instance that returns plain output
    (without color for example)"""
    ui = UI.ui()
    ui.setconfig('extensions', 'color', None)
    return ui


def shell_cmd(root, cmd):
    """Return the output of `cmd` using an external shell."""
    if type(cmd) is tuple:
        cmd = list(cmd)
    elif type(cmd) is not list:
        cmd = [cmd]
    p = Popen([ASCFG.get('hg.bin'), '-R', root] + cmd, stdout=PIPE)
    return p.communicate()[0].strip()

LOGopts = {
    'date': None,  # required by hg 2.2.2
    'user': None,  # required by hg 2.0
}


def get_repo_id(repo, rev='0'):
    """Return the hash of the revision 'rev'."""
    opts = {
        'rev': (rev, ),
        'template':  '{node}\n',
    }
    opts.update(LOGopts)
    try:
        out = hgcmd(newui(), commands.log, repo, **opts)[1]
    except Exception:
        out = None
    return out


def get_repo_name(repo):
    """Return the name of the repository (using its id).
    Returns 'None' if it is not known."""
    return REPOID.get(get_repo_id(repo), None)


def get_repo_by_rev(repo):
    """Return the name of the repository (using its id).
    Returns 'None' if it is not known."""
    for rev, name in REPOID.items():
        if get_repo_id(repo, rev):
            return name
    return None


def has_uncommitted_changes(repo, unknown=False):
    """Tell if the repo has uncommitted changes"""
    st = hgcmd(newui(), commands.status, repo,
               modified=True, added=True, removed=True, deleted=True,
               unknown=unknown)[1]
    return st.strip() != ''


def check_for_updates(path, **opts):
    """Check for available updates for the given repository
    Returns 0 if there are incoming changes, 1 otherwise, 4 in case of error."""
    logger.debug(_("checking for updates for repository '%s'"), path)
    repo = hg.repository(newui(), path)
    opts['bundle'] = False
    opts['force'] = False
    try:
        retcode = commands.incoming(repo.ui, repo, **opts)
    except:
        retcode = 4
    return retcode


def get_changed_files(ui, repo, revs=(), files=()):
    """Return the list of files changed between two revisions"""
    output = hgcmd(ui, commands.status, repo, rev=revs, *files)[1]
    lines = output.splitlines()
    dchg = {}
    add, mod, rmv, mis, unk = [], [], [], [], []
    for occ in lines:
        try:
            fields = occ.split()
            assert len(fields) == 2, occ
            chg, fname = fields
            if chg == 'R':
                rmv.append(fname)
            elif chg == 'A':
                add.append(fname)
            elif chg == 'M':
                mod.append(fname)
            elif chg == '!':
                mis.append(fname)
            elif chg == '?':
                unk.append(fname)
            else:
                # ignore clean & ignored files
                pass
        except IndexError:
            raise ValueError('line: %r' % occ)
    dchg['A'] = add
    dchg['M'] = mod
    dchg['R'] = rmv
    dchg['!'] = mis
    dchg['?'] = unk
    msg = []
    if len(add):
        msg.append(_('%d files added') % len(add))
        logger.debug(_('added files:'))
        logger.debug(green(os.linesep.join(add)))
    if len(mod):
        msg.append(_('%d files changed') % len(mod))
        logger.debug(_('changed files:'))
        logger.debug(magenta(os.linesep.join(mod)))
    if len(rmv):
        msg.append(_('%d files removed') % len(rmv))
        logger.debug(_('removed files:'))
        logger.debug(red(os.linesep.join(rmv)))
    if len(mis):
        msg.append(_('%d files missing') % len(mis))
        logger.debug(_('missing files:'))
        logger.debug(red(os.linesep.join(mis)))
    if len(unk):
        msg.append(_('%d files not tracked') % len(unk))
        logger.debug(_('unknown files:'))
        logger.debug(red(os.linesep.join(unk)))
    dchg['msg'] = ', '.join(msg)
    return dchg


def get_diff(ui, repo, rev, files):
    """Return the diff of files between two revisions"""
    iret, output = hgcmd(ui, commands.diff, repo,
                         rev=rev, ignore_space_change=True, *files)
    if iret not in (0, None):
        logger.info(_("'hg diff' exists with code %s"), iret)
    return output


def get_hgpath(ui, repo, dest):
    """Return the full url of a destination"""
    return hgcmd(ui, commands.paths, repo, dest)[1]


def valid_branch(repo, branch):
    """Return True if 'branch' is one of the open branches
    (default or the last vNN)"""
    if branch == 'default':
        return True
    out = hgcmd(newui(), commands.branches, repo, active=True)[1]
    lbr = [line.split()[0] for line in out.splitlines() if line]
    # keep branches matching 'vNN'
    num = []
    for ibr in lbr:
        mat = re.search('^v([0-9]+)$', ibr)
        if mat:
            num.append(int(mat.group(1)))
    num.sort()
    last = 'v%d' % num[-1]
    return branch == last


def get_node(ui, repo, revset):
    """Return nodes from a revset."""
    logger.debug(_("revset: {0}").format(revset))
    try:
        node = hgcmd(ui, commands.log, repo, rev=(revset, ),
                     template='{node}\n', **LOGopts)[1]
    except Exception:
        raise IOError("invalid revset: '{0}'".format(revset))
    return node


def get_reference_head(ui, repo, branch='.', source=None):
    """Return the current integrated node in the branch"""
    source = source or "codeaster"
    rset = "last(branch({0}) and not outgoing('{1}') and not secret())".format(
        branch, source)
    return get_node(ui, repo, rset)


def get_reference_ancestor(ui, repo, child, branch, source):
    """Return the node of the last reference ancestor of a revision in a
    branch.

    Arguments:
        child (str): Revision of which the ancestor is searched.
        branch (str): Branch name in which the ancestor is searched.
        source (str): Repository path that publish reference revisions.

    Returns:
        str: Revision identifier.
    """
    rset = ("last(branch('{1}') and not outgoing('{2}') and not secret() "
            "and ancestors({0}))").format(child, branch, source)
    return get_node(ui, repo, rset)


def parent_is_last(path, branch="default"):
    """Check that the parent revision is the newest one if the given branch."""
    # not relevant in all repositories
    repo = hg.repository(newui(), path)
    ui = repo.ui
    last = hgcmd(ui, commands.log, repo, template='{node}\n',
                 rev=("last({})".format(branch), ), **LOGopts)[1]
    par = hgcmd(ui, commands.parents, repo, template='{node}\n', **LOGopts)[1]
    return par == last


def get_parent_desc(repodir):
    """Return the short description of the parent revision"""
    cmd = ['hg', 'parent', '-R', repodir,
           '--template', '{node|short}: {desc|firstline}']
    return Popen(cmd, stdout=PIPE, stderr=PIPE).communicate()[0]


def get_solved_issues(ui, repo, rev1, rev2):
    """Return 3 lists: issues solved between the two revisions in all the
    bugtrackers, in rex only, in bitbucket only."""
    revrange = str(rev1) + '::' + str(rev2) + ' and not ' + str(rev1)
    output = hgcmd(ui, commands.log, repo, rev=(revrange, ),
                   template='{desc}\n', **LOGopts)[1]
    ret = []
    for track in BUGTRACKERS:
        ret.append(get_issues_from_descr(output, filter=track))
    return ret


def get_issues_from_descr(text, filter=None, only=False, bracket=True):
    """Extract solved issues numbers from a text.

    Arguments:
        text (str): Text of the issues description.
        filter (str): One of 'rex', 'bitbucket' or None. Limits the
            list of issues to the given bugtracker (default: None).
        only (bool): If *True*, 'text' must only contain issue numbers,
            otherwise it raises *ValueError* (default: False).
        bracket (bool): If *True*, only issues numbers found between brackets
            ('[]') are returned, otherwise they are ignored (default: True).

    Returns;
        list[int]: List of issues numbers in the same order as they are found.
    """
    if bracket:
        brack = re.compile(r'\[(.*?)\]', re.M)
        search_in = brack.findall(text)
    else:
        search_in = [text]
    expr = re.compile(RE_ISSUE, re.M)
    clean = re.compile('[, ]+')
    issues = []
    for enclosed in search_in:
        buff = enclosed
        for mat in expr.finditer(enclosed):
            if filter is None \
                or (filter == 'rex' and mat.group('type') in ('#', 'issue')) \
                or (filter == 'bitbucket'
                    and mat.group('type') in ('bb', 'BB')):
                num = int(mat.group('number'))
                if num not in issues:
                    issues.append(num)
                buff = buff.replace(mat.group('issue'), "")
        buff = clean.sub('', buff)
        if only and buff:
            raise ValueError("unexpected issue number: {0}".format(buff))
    return issues


def get_description(ui, repo, rev1, rev2=None):
    """Return the list of the description of each revision
    between `rev1` and `rev2`"""
    rev2 = rev2 or rev1
    revrange = str(rev1) + '::' + str(rev2)
    output = hgcmd(ui, commands.log, repo, rev=(revrange, ),
                   template='_DELIMITER_{desc}\n', **LOGopts)[1]
    descr = [txt for txt in output.split('_DELIMITER_') if txt]
    return descr


def split_strip(string, sep):
    """Split a string and strip each element"""
    if not string:
        return []
    return [elt.strip() for elt in string.split(sep)]


def sendmail(recipients, subject, text=None, fromaddr=None, footer=True,
             bcc=False, dry_run=False):
    """Send the `text` to `recipients`
    `recipients` is a valid list of email addresses or a role of the
    [aster]/notify section.
    If `text` is None, send the logger content."""
    from email.mime.multipart import MIMEMultipart
    from email.mime.text import MIMEText
    if logger.debug_enabled():
        dry_run = True
    if type(recipients) not in (list, tuple):
        recipients = [recipients]
    logger.debug(_("mail recipients: %r"), recipients)
    toaddr = []
    for addr in recipients:
        if not addr:
            continue
        elif '@' in addr:
            addr = [addr]
        else:
            addr = split_strip(ASCFG.get('notify.' + addr, ''), ',')
        for new in addr:
            if not new in toaddr:
                toaddr.append(new)
    if not toaddr:
        if recipients:
            logger.warn(_("no destination address, %s is neither a valid address "
                          "nor defined in the [aster] section"), recipients)
        return toaddr
    else:
        admsg = toaddr
        if len(toaddr) == 1:
            admsg = toaddr[0]
        logger.info(_("sending email to %s"), admsg)
    if dry_run:
        logger.info(_("(DRY RUN) not sent"))
        return toaddr
    host = os.uname()[1]
    # shown 'from' address
    fromaddr = fromaddr or '%s@%s' % (getpass.getuser(), host)
    # user connected to the mta (may be filtered, localhost should be authorized)
    sender = '%s@localhost' % getpass.getuser()
    mailhost = ASCFG.get('notify.mailhost')
    port = ASCFG.get('notify.port')
    kwargs = {}
    if mailhost:
        kwargs['host'] = mailhost
    if port:
        kwargs['port'] = port
    msg = MIMEMultipart('alternative')
    msg["Subject"] = to_unicode(subject).format(host=host,
                                                date=time.strftime('%Y-%m-%d'))
    msg['From'] = fromaddr
    if not bcc:
        msg['To'] = ', '.join(toaddr)
    else:
        msg['To'] = fromaddr

    if text is None:
        text = logger.get_text()
    if footer:
        text = to_unicode(text) + os.linesep * \
            2 + to_unicode(_("sent from %s") % host)
    part1 = MIMEText(to_unicode(text), "plain", "utf-8")
    msg.attach(part1)
    try:
        smtp = smtplib.SMTP(**kwargs)
        smtp.connect()
        smtp.sendmail(sender, toaddr, msg.as_string())
        smtp.quit()
    except Exception, exc:
        logger.debug(_("can not send the email:\nException: %s"), str(exc))
        logger.debug(_("this happens on some machines."))
        logger.debug(_("trying using the shell command..."))
        sendmail_shell(msg, text)
    return toaddr


def sendmail_shell(msg, text):
    """Send a mail using mailutils shell commands"""
    import tempfile
    try:
        tmpname = tempfile.NamedTemporaryFile().name
        tmpf = open(tmpname, 'wb')
        tmpf.write(convert(text))
        tmpf.close()
        pipe = open(tmpname, 'rb')
        cmd = ['mail', '-s', convert(msg["Subject"]), msg['To']]
        logger.debug('trying: %s', cmd)
        pmail = Popen(cmd, stdin=pipe)
        pipe.close()
        pmail.wait()
        logger.debug(_("mail sent using the shell command"))
    except Exception, exc:
        logger.warn(
            _("can not send the email using mailutils:\nException: %s"),
            str(exc))


def colorize_status(status):
    """Colorize the status"""
    ident = lambda x: x
    dfunc = {'M': magenta, '?': blue, '!': cyan,
             'A': green, 'R': red, 'I': grey}
    new = []
    for line in status.splitlines():
        mat = re.search('^([MAR\?!]) ', line)
        if mat:
            line = dfunc.get(mat.group(1), ident)(line)
        new.append(line)
    return os.linesep.join(new)

_author_regexp = re.compile(" *(.*?) *< *(.*?) *> *")

def decode_author(author):
    """Extract author name and email"""
    mat = _author_regexp.search(author)
    if not mat:
        logger.warn(_("invalid author field: {}").format(author))
        return author, ""
    return mat.groups()
